/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include <SDL.h>
#include <SDL_opengl.h>
#include "render.h"
#include "scaler.h"

#define OSB_COPPER  0
#define OSB_SPRITES 1

#define kTileW 16
#define kTileH 16
#define kTilesCount 256

static const int kDefaultScaler = 4;

typedef struct {
	void (*proc)(uint16 *dst, int dstPitch, const uint16 *src, int srcPitch, int w, int h);
	int factor;
} scaler_desc_t;

struct Texture {
	GLuint id;
	int w, h;
	GLfloat u, v;
};

struct TileTexture {
	GLuint id;
	int w, h;
	GLfloat u, v;
};

struct SpriteTexture {
	const uint8_t *data;
	Texture tex[256];
};

static Texture _backgroundTex;
static TileTexture _tilesTex;
static SpriteTexture _cacheTex[16];

static const scaler_desc_t _scalers[] = {
	{ point1x, 1 },
	{ point2x, 2 },
	{ scale2x, 2 },
	{ point3x, 3 },
	{ scale3x, 3 },
};

static int roundPow2(int x) {
	if ((x & (x - 1)) == 0) {
		return x;
	} else {
		int num = 1;
		while (num < x) {
			num <<= 1;
		}
		return num;
	}
}

void Render::init(int w, int h) {
	_w = w;
	_h = h;

	static bool inited = false;
	if (inited) {
		return;
	}
	inited = true;

	memset(_tex8to5551, 0, sizeof(_tex8to5551));

	_scaler = kDefaultScaler;
	_texBuf = (uint16 *)malloc(512 * 256 * sizeof(uint16));
	_backgroundTex.id = -1;
	_tilesTex.id = -1;
	memset(_cacheTex, 0, sizeof(_cacheTex));
	memset(_spritesTex, 0, sizeof(_spritesTex));

	memset(offscreen_buffers_table, 0, sizeof(offscreen_buffers_table));
	const int buffer_w = 320 * _scalers[_scaler].factor;
	const int buffer_h = 160 * _scalers[_scaler].factor;
	for (int i = 0; i < (int)ARRAYSIZE(offscreen_buffers_table); ++i) {
		offscreen_buffer_t *ob = &offscreen_buffers_table[i];
		ob->pitch = buffer_w;
		ob->size = buffer_w * buffer_h;
		ob->ptr = (uint8 *)malloc(ob->size);
		if (!ob->ptr) {
			error("Unable to allocate buffer %d (size %d)", i, ob->size);
		}
	}
	_tilesBuffer = (uint8 *)malloc(kTilesCount * kTileH * kTileW);
	if (!_tilesBuffer) {
		error("Unable to allocate tiles buffer");
	}
}

void Render::destroy() {
	for (int i = 0; i < (int)ARRAYSIZE(offscreen_buffers_table); ++i) {
		free(offscreen_buffers_table[i].ptr);
		offscreen_buffers_table[i].ptr = 0;
	}
	free(_tilesBuffer);
}

void Render::convertTexture(uint16 *dst, int dstPitch, const uint8 *src, int srcPitch, int w, int h, const uint16 *lut) {
	for (int y = 0; y < h; ++y) {
		for (int x = 0; x < w; ++x) {
			const uint8_t color = src[y * srcPitch + x];
			_texBuf[y * w + x] = lut[color];
		}
	}
	_scalers[_scaler].proc(dst, dstPitch, _texBuf, w, w, h);
}

void Render::uploadSpriteTexture(const uint8 *data, int pal, SpriteTexture *cacheTex) {
	cacheTex->data = data;
	const int num = READ_LE_UINT16(data) / 2;
	for (int i = 0; i < num; ++i) {
		const uint8 *ptr = data + READ_LE_UINT16(data + i * 2);
		const int w = READ_LE_UINT16(ptr); ptr += 2;
		const int h = READ_LE_UINT16(ptr); ptr += 2;
		if (w == 0 || h == 0) {
			continue;
		}
		const int dstPitch = offscreen_buffers_table[OSB_SPRITES].pitch;
		memset(offscreen_buffers_table[OSB_SPRITES].ptr, 0, dstPitch * h); // TEMP
		drawSprite(ptr, 0, 0, w, h, 0, pal, offscreen_buffers_table[OSB_SPRITES].ptr, dstPitch);
		Texture *t = &cacheTex->tex[i];
		t->w = w * _scalers[_scaler].factor;
		t->h = h * _scalers[_scaler].factor;
		const int texW = roundPow2(t->w);
		const int texH = roundPow2(t->h);
		t->u = t->w / (GLfloat)texW;
		t->v = t->h / (GLfloat)texH;
		glGenTextures(1, &t->id);
		glBindTexture(GL_TEXTURE_2D, t->id);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		uint16 *buf = (uint16 *)malloc(texW * texH * sizeof(uint16_t));
		if (buf) {
			convertTexture(buf, texW, offscreen_buffers_table[OSB_SPRITES].ptr, dstPitch, w, h, _tex8to5551);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texW, texH, 0, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, buf);
			free(buf);
		}
	}
}

void Render::flushTextures() {
	if (_backgroundTex.id != -1) {
		glDeleteTextures(1, &_backgroundTex.id);
		_backgroundTex.id = -1;
	}
	if (_tilesTex.id != -1) {
		glDeleteTextures(1, &_tilesTex.id);
		_tilesTex.id = -1;
	}
	for (int i = 0; i < 16; ++i) {
		if (_cacheTex[i].data) {
			const int num = READ_LE_UINT16(_cacheTex[i].data);
			for (int j = 0; j < num; ++j) {
				Texture *t = &_cacheTex[i].tex[j];
				if (t->w != 0 && t->h != 0) {
					glDeleteTextures(1, &t->id);
				}
			}
		}
	}
	memset(_cacheTex, 0, sizeof(_cacheTex));
	memset(_spritesTex, 0, sizeof(_spritesTex));
}

static int convertColor5551(const int color) {
	const int r = ((color >> 8) & 15) << 1;
	const int g = ((color >> 4) & 15) << 1;
	const int b = ( color       & 15) << 1;
	return (r << 11) | (g << 6) | (b << 1) | 1;
}

void Render::setPalette(const uint16 *pal_data, int num_colors, int num) {
	if (num == 0) {
		_tex8to5551[0] = 0;
		for (int i = 1; i < num_colors; ++i) {
			_tex8to5551[i] = convertColor5551(pal_data[i]);
		}
		_tex8to5551[240] = 0;
		_tex8to5551[241] = 0xFFFF;
		_tex8to5551[242] = 1;
	} else {
		for (int i = 0; i < num_colors; ++i) {
			_tex8to5551[num * 16 + i] = convertColor5551(pal_data[i]);
		}
	}
}

static void draw_img_16(const uint8 *src, const uint8 *lut, int pitch, int x, int y, int w, int h, uint8 *dst, int dstPitch, int dstW, int dstH) {
	w = (w + 7) / 8;
	for (int j = 0; j < h; ++j) {
		const int cy = y + j;
		for (int i = 0; i < w; ++i) {
			for (int bit = 0; bit < 8; ++bit) {
				const int cx = x + i * 8 + bit;
				if (cx >= 0 && cx < dstW && cy >= 0 && cy < dstH) {
					const int mask = 1 << (7 - bit);
					uint8 color = 0;
					for (int plane = 0; plane < 4; ++plane) {
						if (src[plane * w + i] & mask) {
							color |= 1 << plane;
						}
					}
					color = lut[color];
					if (color != 8) {
						dst[cy * dstPitch + cx] = color;
					}
				}
			}
		}
		src += pitch;
	}
}

void Render::drawSprite(const uint8 *spr_data, int x, int y, int w, int h, int flags, int pal_num, uint8 *dst, int dstPitch) {
	uint8 data[4];
	for (int j = 0; j < h; ++j) {
		const int cy = y + j;
		for (int i = 0; i < (w + 7) / 8; ++i) {
			memcpy(data, spr_data, 4);
			spr_data += 4;
			for (int b = 0; b < 8; ++b) {
				const int cx = x + i * 8 + b;
				if (cx >= 0 && cx < _w && cy >= 0 && cy < _h) {
					const int mask = 1 << (7 - b);
					uint8 color = 0;
					for (int p = 0; p < 4; ++p) {
						if (data[p] & mask) {
							color |= 1 << p;
						}
					}
					if (color != 0) {
						color += pal_num * 16;
						dst[cx] = color;
					}
				}
			}
		}
		dst += dstPitch;
	}
}

static void emitQuadTex(int x1, int y1, int x2, int y2, GLfloat u1, GLfloat v1, GLfloat u2, GLfloat v2) {
	glBegin(GL_QUADS);
		glTexCoord2f(u1, v1);
		glVertex2i(x1, y1);
		glTexCoord2f(u2, v1);
		glVertex2i(x2, y1);
		glTexCoord2f(u2, v2);
		glVertex2i(x2, y2);
		glTexCoord2f(u1, v2);
		glVertex2i(x1, y2);
	glEnd();
}

void Render::updateScreen() {
	glEnable(GL_ALPHA_TEST);
	glEnable(GL_TEXTURE_2D);
	glViewport(0, 0, _w, _h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, _w, _h, 0, 0, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glPushMatrix();
	static const int kW = 320;
	static const int kH = 160;
	glScalef(_w / (GLfloat)kW, _h / (GLfloat)kH, 1.);
//	glClear(GL_COLOR_BUFFER_BIT);
	glAlphaFunc(GL_ALWAYS, 1);
	// copper+background
	if (_backgroundTex.id != -1) {
		glBindTexture(GL_TEXTURE_2D, _backgroundTex.id);
		emitQuadTex(0, 0, kW, kH, 0, 0, _backgroundTex.u, _backgroundTex.v);
	}
	glAlphaFunc(GL_EQUAL, 1);
	// tilemap
	if (_tilesTex.id != -1) {
		glBindTexture(GL_TEXTURE_2D, _tilesTex.id);
		int x = -(_tileMapOriginX & 15);
		int y = -(_tileMapOriginY & 15);
		const int tileMapDx = _tileMapOriginX >> 4;
		const int tileMapDy = _tileMapOriginY >> 4;
		for (int j = 0; j <= kH / 16; ++j) {
			for (int i = 0; i <= kW / 16; ++i) {
				int num = _tileMapData[(tileMapDx + i) * _tileMapH + (tileMapDy + j)];
				num = _tileMapAnimLookupTable[num];
				if (num != 0) {
					const GLfloat u1 = _tilesTex.u *  (num & 15)      / 16.;
					const GLfloat u2 = _tilesTex.u * ((num & 15) + 1) / 16.;
					const GLfloat v1 = _tilesTex.v *  (num >> 4)      / 16.;
					const GLfloat v2 = _tilesTex.v * ((num >> 4) + 1) / 16.;
					emitQuadTex(x + i * 16, y, x + i * 16 + 16, y + 16, u1, v1, u2, v2);
				}
			}
			y += 16;
		}
	}
	// sprites
	for (int i = 0; i < _spritesListCount; ++i) {
		sprite_t *spr = &_spritesList[i];
		if (_spritesTex[i]) {
			const uint8 *spr_data = spr->data + READ_LE_UINT16(spr->data + spr->num * 2);
			const int w = READ_LE_UINT16(spr_data); spr_data += 2;
			const int h = READ_LE_UINT16(spr_data); spr_data += 2;
			glBindTexture(GL_TEXTURE_2D, _spritesTex[i]->id);
			emitQuadTex(spr->x, spr->y, spr->x + w, spr->y + h, 0., 0., _spritesTex[i]->u, _spritesTex[i]->v);
		}
	}
	// dump palette
if (0) {
	glDisable(GL_TEXTURE_2D);
	for (int i = 0; i < 16; ++i) {
		const GLfloat r = ((_tex8to5551[i] >> 11) & 31) / (GLfloat)31;
		const GLfloat g = ((_tex8to5551[i] >>  6) & 31) / (GLfloat)31;
		const GLfloat b = ((_tex8to5551[i] >>  1) & 31) / (GLfloat)31;
		glColor4f(r, g, b, 1.);
		glBegin(GL_QUADS);
			const int x1 = i * 8;
			const int x2 = x1 + 8;
			glVertex2i(x1, 0);
			glVertex2i(x2, 0);
			glVertex2i(x2, 8);
			glVertex2i(x1, 8);
		glEnd();
	}
}
	glColor4f(1., 1., 1., 1.);
	glPopMatrix();
}

static void decodeTile(uint8 *dst, const uint8 *src) {
	uint16 data[4];
	for (int y = 0; y < 16; ++y) {
		for (int plane = 0; plane < 4; ++plane) {
			data[plane] = READ_BE_UINT16(src);
			src += 2;
		}
		for (int bit = 0; bit < 16; ++bit) {
			const int mask = 1 << (15 - bit);
			uint8 color = 0;
			for (int plane = 0; plane < 4; ++plane) {
				if (data[plane] & mask) {
					color |= 1 << plane;
				}
			}
			*dst++ = color;
		}
	}
}

void Render::setTileData(const uint8 *src, int transparentColor) {
	for (int i = 0; i < kTilesCount; ++i) {
		uint8 buf[16 * 16];
		decodeTile(buf, src);
		src += 16 * 8;
		if (transparentColor != 0) {
			for (int i = 0; i < 16 * 16; ++i) {
				if (buf[i] == transparentColor) {
					buf[i] = 0;
				}
			}
		}
		const int xOffset = i * 16;
		for (int y = 0; y < 16; ++y) {
			memcpy(_tilesBuffer + xOffset + y * 16 * kTilesCount, buf + y * 16, 16);
		}
	}
	const int w = 16 * (kTilesCount / 16);
	const int h = 16 * (kTilesCount / 16);
	glGenTextures(1, &_tilesTex.id);
	_tilesTex.w = w * _scalers[_scaler].factor;
	_tilesTex.h = h * _scalers[_scaler].factor;
	const int texW = roundPow2(_tilesTex.w);
	const int texH = roundPow2(_tilesTex.h);
	_tilesTex.u = _tilesTex.w / (GLfloat)texW;
	_tilesTex.v = _tilesTex.h / (GLfloat)texH;
	uint16 *buf = (uint16 *)malloc(texW * texH * sizeof(uint16_t));
	if (buf) {
		for (int y = 0; y < 16; ++y) {
			for (int x = 0; x < 16; ++x) {
				int i = y * 16 + x;
				convertTexture(buf + (y * texW + x) * 16 * _scalers[_scaler].factor, texW, _tilesBuffer + i * 16, 16 * kTilesCount, 16, 16, _tex8to5551);
			}
		}
		glBindTexture(GL_TEXTURE_2D, _tilesTex.id);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texW, texH, 0, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, buf);
		free(buf);
	}
}

void Render::setTileMapData(const uint8 *data, int w, int h, const uint8 *anim_lut) {
	_tileMapData = data;
	_tileMapAnimLookupTable = anim_lut;
	_tileMapOriginX = 0;
	_tileMapOriginY = 0;
	_tileMapW = w;
	_tileMapH = h;
}

void Render::setTileMapOrigin(int x, int y) {
	_tileMapOriginX = x;
	_tileMapOriginY = y;
}

void Render::setCopper(const uint16 *pal_data, int num_lines) {
	const int colorOffset = 240 - num_lines;
	for (int i = 0; i < num_lines; ++i) {
		_tex8to5551[colorOffset + i] = convertColor5551(pal_data[i]);
	}
	uint8 *copper_buf = offscreen_buffers_table[OSB_COPPER].ptr;
	for (int i = 0; i < _h; ++i) {
		memset(copper_buf + i * 320, colorOffset + i * num_lines / 160, 320);
	}
}

void Render::setBackgroundFrame(const uint8 *frame_data, int w, int h, const uint8 *lut) {
	draw_img_16(frame_data, lut, 320, 0, (160 - 32) / 2, 640, 32, offscreen_buffers_table[OSB_COPPER].ptr, 320, 320, 160);
	w = 320;
	h = 160;
	_backgroundTex.w = w * _scalers[_scaler].factor;
	_backgroundTex.h = h * _scalers[_scaler].factor;
	const int texW = roundPow2(_backgroundTex.w);
	const int texH = roundPow2(_backgroundTex.h);
	_backgroundTex.u = _backgroundTex.w / (GLfloat)texW;
	_backgroundTex.v = _backgroundTex.h / (GLfloat)texH;
	glGenTextures(1, &_backgroundTex.id);
	glBindTexture(GL_TEXTURE_2D, _backgroundTex.id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	uint16 *buf = (uint16 *)malloc(texW * texH * sizeof(uint16_t));
	if (buf) {
		convertTexture(buf, texW, offscreen_buffers_table[OSB_COPPER].ptr, w, w, h, _tex8to5551);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texW, texH, 0, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, buf);
		free(buf);
	}
}

void Render::clearSpriteList() {
	_spritesListCount = 0;
}

void Render::addToSpriteList(const uint8 *data, int num, int x, int y, int pal) {
	if (_spritesListCount < kMaxSpritesCount) {
		sprite_t *spr = &_spritesList[_spritesListCount];
		spr->data = data;
		spr->num = num;
		spr->x = x;
		spr->y = y;
		spr->flags = 0;
		spr->pal_num = pal;
		int index = -1;
		for (int i = 0; i < 16; ++i) {
			if (_cacheTex[i].data == data) {
				_spritesTex[_spritesListCount] = &_cacheTex[i].tex[num];
				++_spritesListCount;
				return;
			} else if (_cacheTex[i].data == 0) {
				index = i;
			}
		}
		if (index == -1) {
			warning("addToSpriteList() no free index");
			++_spritesListCount;
			return;
		}
		uploadSpriteTexture(data, pal, &_cacheTex[index]);
		_spritesTex[_spritesListCount] = &_cacheTex[index].tex[num];
		++_spritesListCount;
	} else {
		warning("addToSpriteList() _spritesList overflow");
	}
}

